@@ -27,16 +27,20 @@
 #include "searchresult.h"
 
 #include <util/plist.h>
+#include <util/sqlitedatabase.h>
 
 #include <QDir>
 #include <QFile>
+#include <QJsonArray>
 #include <QJsonDocument>
 #include <QJsonObject>
 #include <QRegularExpression>
-#include <QSqlDatabase>
-#include <QSqlError>
-#include <QSqlQuery>
 #include <QVariant>
+#include <QVarLengthArray>
+
+#include <sqlite3.h>
+
+#include <cstring>
 
 using namespace Zeal::Registry;
 
@@ -57,6 +61,8 @@ const char DocSetPlatformFamily[] = "DocSetPlatformFamily";
 }
 }
 
+static void sqliteScoreFunction(sqlite3_context *context, int argc, sqlite3_value **argv);
+
 Docset::Docset(const QString &path) :
     m_path(path)
 {
@@ -68,7 +74,7 @@ Docset::Docset(const QString &path) :
 
     // Attempt to find the icon in any supported format
     for (const QString &iconFile : dir.entryList({QStringLiteral("icon.*")}, QDir::Files)) {
-        m_icon = QIcon(dir.absoluteFilePath(iconFile));
+        m_icon = QIcon(dir.filePath(iconFile));
         if (!m_icon.availableSizes().isEmpty())
             break;
     }
@@ -82,9 +88,9 @@ Docset::Docset(const QString &path) :
     // /Articles/ConfigFiles.html
     Util::Plist plist;
     if (dir.exists(QStringLiteral("Info.plist")))
-        plist.read(dir.absoluteFilePath(QStringLiteral("Info.plist")));
+        plist.read(dir.filePath(QStringLiteral("Info.plist")));
     else if (dir.exists(QStringLiteral("info.plist")))
-        plist.read(dir.absoluteFilePath(QStringLiteral("info.plist")));
+        plist.read(dir.filePath(QStringLiteral("info.plist")));
     else
         return;
 
@@ -93,7 +99,7 @@ Docset::Docset(const QString &path) :
 
     if (m_name.isEmpty()) {
         // Fallback if meta.json is absent
-        if (!plist.contains(InfoPlist::CFBundleName)) {
+        if (plist.contains(InfoPlist::CFBundleName)) {
             m_name = m_title = plist[InfoPlist::CFBundleName].toString();
             // TODO: Remove when MainWindow::docsetName() will not use directory name
             m_name.replace(QLatin1Char(' '), QLatin1Char('_'));
@@ -116,18 +122,24 @@ Docset::Docset(const QString &path) :
     if (!dir.cd(QStringLiteral("Resources")) || !dir.exists(QStringLiteral("docSet.dsidx")))
         return;
 
-    QSqlDatabase db = QSqlDatabase::addDatabase(QStringLiteral("QSQLITE"), m_name);
-    db.setDatabaseName(dir.absoluteFilePath(QStringLiteral("docSet.dsidx")));
+    m_db = new Util::SQLiteDatabase(dir.filePath(QStringLiteral("docSet.dsidx")));
 
-    if (!db.open()) {
-        qWarning("SQL Error: %s", qPrintable(db.lastError().text()));
+    if (!m_db->isOpen()) {
+        qWarning("SQL Error: %s", qPrintable(m_db->lastError()));
         return;
     }
 
-    m_type = db.tables().contains(QStringLiteral("searchIndex")) ? Type::Dash : Type::ZDash;
+    sqlite3_create_function(m_db->handle(), "zealScore", 2, SQLITE_UTF8, nullptr,
+                            sqliteScoreFunction, nullptr, nullptr);
+
+    m_type = m_db->tables().contains(QStringLiteral("searchIndex")) ? Type::Dash : Type::ZDash;
 
     createIndex();
 
+    if (m_type == Docset::Type::ZDash) {
+        createView();
+    }
+
     if (!dir.cd(QStringLiteral("Documents"))) {
         m_type = Type::Invalid;
         return;
@@ -166,7 +178,7 @@ Docset::Docset(const QString &path) :
 
 Docset::~Docset()
 {
-    QSqlDatabase::removeDatabase(m_name);
+    delete m_db;
 }
 
 bool Docset::isValid() const
@@ -206,7 +218,7 @@ QString Docset::path() const
 
 QString Docset::documentPath() const
 {
-    return QDir(m_path).absoluteFilePath(QStringLiteral("Contents/Resources/Documents"));
+    return QDir(m_path).filePath(QStringLiteral("Contents/Resources/Documents"));
 }
 
 QIcon Docset::icon() const
@@ -246,45 +258,46 @@ const QMap<QString, QUrl> &Docset::symbols(const QString &symbolType) const
 
 QList<SearchResult> Docset::search(const QString &query, const CancellationToken &token) const
 {
-    // Make it safe to use in a SQL query.
-    QString sanitizedQuery = query;
-    sanitizedQuery.replace(QLatin1String("\\"), QLatin1String("\\\\"));
-    sanitizedQuery.replace(QLatin1String("_"), QLatin1String("\\_"));
-    sanitizedQuery.replace(QLatin1String("%"), QLatin1String("\\%"));
-    sanitizedQuery.replace(QLatin1String("'"), QLatin1String("''"));
-
-    QString queryStr;
+    QString sql;
     if (m_type == Docset::Type::Dash) {
-        queryStr = QStringLiteral("SELECT name, type, path "
-                                  "    FROM searchIndex "
-                                  "WHERE (name LIKE '%%1%' ESCAPE '\\') "
-                                  "ORDER BY name COLLATE NOCASE").arg(sanitizedQuery);
+        if (m_fuzzySearchEnabled) {
+            sql = QStringLiteral("SELECT name, type, path, '', zealScore('%1', name) as score"
+                                 "  FROM searchIndex"
+                                 "  WHERE score > 0");
+        } else {
+            sql = QStringLiteral("SELECT name, type, path, ''"
+                                 "  FROM searchIndex"
+                                 "  WHERE (name LIKE '%%1%' ESCAPE '\\')");
+        }
     } else {
-        queryStr = QStringLiteral("SELECT ztokenname, ztypename, zpath, zanchor "
-                                  "    FROM ztoken "
-                                  "LEFT JOIN ztokenmetainformation "
-                                  "    ON ztoken.zmetainformation = ztokenmetainformation.z_pk "
-                                  "LEFT JOIN zfilepath "
-                                  "    ON ztokenmetainformation.zfile = zfilepath.z_pk "
-                                  "LEFT JOIN ztokentype "
-                                  "    ON ztoken.ztokentype = ztokentype.z_pk "
-                                  "WHERE (ztokenname LIKE '%%1%' ESCAPE '\\') "
-                                  "ORDER BY ztokenname COLLATE NOCASE").arg(sanitizedQuery);
+        if (m_fuzzySearchEnabled) {
+            sql = QStringLiteral("SELECT name, type, path, fragment, zealScore('%1', name) as score"
+                                 "  FROM searchIndex"
+                                 "  WHERE score > 0");
+        } else {
+            sql = QStringLiteral("SELECT name, type, path, fragment"
+                                 "  FROM searchIndex"
+                                 "  WHERE (name LIKE '%%1%' ESCAPE '\\')");
+        }
     }
 
     // Limit for very short queries.
     // TODO: Show a notification about the reduced result set.
-    if (query.size() < 3)
-        queryStr += QLatin1String(" LIMIT 1000");
+    if (query.size() < 3) {
+        sql += QLatin1String("  LIMIT 1000");
+    }
+
+    // Make it safe to use in a SQL query.
+    QString sanitizedQuery = query;
+    sanitizedQuery.replace(QLatin1Char('\''), QLatin1String("''"));
+    m_db->prepare(sql.arg(sanitizedQuery));
 
     QList<SearchResult> results;
-
-    QSqlQuery sqlQuery(queryStr, database());
-    while (sqlQuery.next() && !token.isCanceled()) {
-        results.append({sqlQuery.value(0).toString(),
-                        parseSymbolType(sqlQuery.value(1).toString()),
-                        const_cast<Docset *>(this),
-                        createPageUrl(sqlQuery.value(2).toString(), sqlQuery.value(3).toString())});
+    while (m_db->next() && !token.isCanceled()) {
+        results.append({m_db->value(0).toString(),
+                        parseSymbolType(m_db->value(1).toString()),
+                        m_db->value(2).toString(), m_db->value(3).toString(),
+                        const_cast<Docset *>(this), m_db->value(4).toInt()});
     }
 
     return results;
@@ -305,25 +318,23 @@ QList<SearchResult> Docset::relatedLinks(const QUrl &url) const
     cleanUrl.setFragment(QString());
 
     // Prepare the query to look up all pages with the same url.
-    QString queryStr;
+    QString sql;
     if (m_type == Docset::Type::Dash) {
-        queryStr = QStringLiteral("SELECT name, type, path FROM searchIndex "
-                                  "WHERE path LIKE \"%1%%\" AND path <> \"%1\"");
+        sql = QStringLiteral("SELECT name, type, path"
+                             "  FROM searchIndex"
+                             "  WHERE path LIKE \"%1%%\" AND path <> \"%1\"");
     } else if (m_type == Docset::Type::ZDash) {
-        queryStr = QStringLiteral("SELECT ztoken.ztokenname, ztokentype.ztypename, zfilepath.zpath, ztokenmetainformation.zanchor "
-                                  "FROM ztoken "
-                                  "LEFT JOIN ztokenmetainformation ON ztoken.zmetainformation = ztokenmetainformation.z_pk "
-                                  "LEFT JOIN zfilepath ON ztokenmetainformation.zfile = zfilepath.z_pk "
-                                  "LEFT JOIN ztokentype ON ztoken.ztokentype = ztokentype.z_pk "
-                                  "WHERE zfilepath.zpath = \"%1\" AND ztokenmetainformation.zanchor IS NOT NULL");
+        sql = QStringLiteral("SELECT name, type, path, fragment"
+                             "  FROM searchIndex"
+                             "  WHERE path = \"%1\" AND fragment IS NOT NULL");
     }
 
-    QSqlQuery sqlQuery(queryStr.arg(cleanUrl.toString()), database());
-    while (sqlQuery.next()) {
-        results.append({sqlQuery.value(0).toString(),
-                        parseSymbolType(sqlQuery.value(1).toString()),
-                        const_cast<Docset *>(this),
-                        createPageUrl(sqlQuery.value(2).toString(), sqlQuery.value(3).toString())});
+    m_db->prepare(sql.arg(cleanUrl.toString()));
+    while (m_db->next()) {
+        results.append({m_db->value(0).toString(),
+                        parseSymbolType(m_db->value(1).toString()),
+                        m_db->value(2).toString(), m_db->value(3).toString(),
+                        const_cast<Docset *>(this), 0});
     }
 
     if (results.size() == 1)
@@ -332,9 +343,9 @@ QList<SearchResult> Docset::relatedLinks(const QUrl &url) const
     return results;
 }
 
-QSqlDatabase Docset::database() const
+QUrl Docset::searchResultUrl(const SearchResult &result) const
 {
-    return QSqlDatabase::database(m_name, true);
+    return createPageUrl(result.urlPath, result.urlFragment);
 }
 
 void Docset::loadMetadata()
@@ -362,31 +373,33 @@ void Docset::loadMetadata()
 
     if (jsonObject.contains(QStringLiteral("extra"))) {
         const QJsonObject extra = jsonObject[QStringLiteral("extra")].toObject();
-        m_indexFileUrl = createPageUrl(extra[QStringLiteral("indexFilePath")].toString());
+
+        if (extra.contains(QStringLiteral("indexFilePath"))) {
+            m_indexFileUrl = createPageUrl(extra[QStringLiteral("indexFilePath")].toString());
+        }
+
+        if (extra.contains(QStringLiteral("keywords"))) {
+            for (const QJsonValueRef kw : extra[QStringLiteral("keywords")].toArray())
+                m_keywords << kw.toString();
+        }
     }
 }
 
 void Docset::countSymbols()
 {
-    QString queryStr;
-    if (m_type == Docset::Type::Dash) {
-        queryStr = QStringLiteral("SELECT type, COUNT(*) FROM searchIndex GROUP BY type");
-    } else if (m_type == Docset::Type::ZDash) {
-        queryStr = QStringLiteral("SELECT ztypename, COUNT(*) FROM ztoken LEFT JOIN ztokentype"
-                                  " ON ztoken.ztokentype = ztokentype.z_pk GROUP BY ztypename");
-    }
-
-    QSqlQuery query(queryStr, database());
-    if (query.lastError().type() != QSqlError::NoError) {
-        qWarning("SQL Error: %s", qPrintable(query.lastError().text()));
+    static const QString sql = QStringLiteral("SELECT type, COUNT(*)"
+                                              "  FROM searchIndex"
+                                              "  GROUP BY type");
+    if (!m_db->prepare(sql)) {
+        qWarning("SQL Error: %s", qPrintable(m_db->lastError()));
         return;
     }
 
-    while (query.next()) {
-        const QString symbolTypeStr = query.value(0).toString();
+    while (m_db->next()) {
+        const QString symbolTypeStr = m_db->value(0).toString();
         const QString symbolType = parseSymbolType(symbolTypeStr);
         m_symbolStrings.insertMulti(symbolType, symbolTypeStr);
-        m_symbolCounts[symbolType] += query.value(1).toInt();
+        m_symbolCounts[symbolType] += m_db->value(1).toInt();
     }
 }
 
@@ -399,28 +412,28 @@ void Docset::loadSymbols(const QString &symbolType) const
 
 void Docset::loadSymbols(const QString &symbolType, const QString &symbolString) const
 {
-    QString queryStr;
+    QString sql;
     if (m_type == Docset::Type::Dash) {
-        queryStr = QStringLiteral("SELECT name, path FROM searchIndex WHERE type='%1' ORDER BY name ASC");
+        sql = QStringLiteral("SELECT name, path"
+                             "  FROM searchIndex"
+                             "  WHERE type='%1'"
+                             "  ORDER BY name");
     } else {
-        queryStr = QStringLiteral("SELECT ztokenname, zpath, zanchor "
-                                  "FROM ztoken "
-                                  "LEFT JOIN ztokenmetainformation ON ztoken.zmetainformation = ztokenmetainformation.z_pk "
-                                  "LEFT JOIN zfilepath ON ztokenmetainformation.zfile = zfilepath.z_pk "
-                                  "LEFT JOIN ztokentype ON ztoken.ztokentype = ztokentype.z_pk WHERE ztypename='%1' "
-                                  "ORDER BY ztokenname ASC");
+        sql = QStringLiteral("SELECT name, path, fragment"
+                             "  FROM searchIndex"
+                             "  WHERE type='%1'"
+                             "  ORDER BY name");
     }
 
-    QSqlQuery query(queryStr.arg(symbolString), database());
-    if (query.lastError().type() != QSqlError::NoError) {
-        qWarning("SQL Error: %s", qPrintable(query.lastError().text()));
+    if (!m_db->prepare(sql.arg(symbolString))) {
+        qWarning("SQL Error: %s", qPrintable(m_db->lastError()));
         return;
     }
 
     QMap<QString, QUrl> &symbols = m_symbols[symbolType];
-    while (query.next())
-        symbols.insertMulti(query.value(0).toString(),
-                            createPageUrl(query.value(1).toString(), query.value(2).toString()));
+    while (m_db->next())
+        symbols.insertMulti(m_db->value(0).toString(),
+                            createPageUrl(m_db->value(1).toString(), m_db->value(2).toString()));
 }
 
 void Docset::createIndex()
@@ -430,19 +443,17 @@ void Docset::createIndex()
     static const QString indexCreateQuery = QStringLiteral("CREATE INDEX IF NOT EXISTS %1%2"
                                                            " ON %3 (%4 COLLATE NOCASE)");
 
-    QSqlQuery query(database());
-
     const QString tableName = m_type == Type::Dash ? QStringLiteral("searchIndex")
                                                    : QStringLiteral("ztoken");
     const QString columnName = m_type == Type::Dash ? QStringLiteral("name")
                                                     : QStringLiteral("ztokenname");
 
-    query.exec(indexListQuery.arg(tableName));
+    m_db->prepare(indexListQuery.arg(tableName));
 
     QStringList oldIndexes;
 
-    while (query.next()) {
-        const QString indexName = query.value(1).toString();
+    while (m_db->next()) {
+        const QString indexName = m_db->value(1).toString();
         if (!indexName.startsWith(IndexNamePrefix))
             continue;
 
@@ -454,9 +465,29 @@ void Docset::createIndex()
 
     // Drop old indexes
     for (const QString &oldIndexName : oldIndexes)
-        query.exec(indexDropQuery.arg(oldIndexName));
+        m_db->execute(indexDropQuery.arg(oldIndexName));
 
-    query.exec(indexCreateQuery.arg(IndexNamePrefix, IndexNameVersion, tableName, columnName));
+    m_db->execute(indexCreateQuery.arg(IndexNamePrefix, IndexNameVersion, tableName, columnName));
+}
+
+void Docset::createView()
+{
+    static const QString viewCreateQuery
+            = QStringLiteral("CREATE VIEW IF NOT EXISTS searchIndex AS"
+                             "  SELECT"
+                             "    ztokenname AS name,"
+                             "    ztypename AS type,"
+                             "    zpath AS path,"
+                             "    zanchor AS fragment"
+                             "  FROM ztoken"
+                             "  INNER JOIN ztokenmetainformation"
+                             "    ON ztoken.zmetainformation = ztokenmetainformation.z_pk"
+                             "  INNER JOIN zfilepath"
+                             "    ON ztokenmetainformation.zfile = zfilepath.z_pk"
+                             "  INNER JOIN ztokentype"
+                             "    ON ztoken.ztokentype = ztokentype.z_pk");
+
+    m_db->execute(viewCreateQuery);
 }
 
 QUrl Docset::createPageUrl(const QString &path, const QString &fragment) const
@@ -478,12 +509,14 @@ QUrl Docset::createPageUrl(const QString &path, const QString &fragment) const
     realPath.remove(dashEntryRegExp);
     realFragment.remove(dashEntryRegExp);
 
-    QUrl url = QUrl::fromLocalFile(QDir(documentPath()).absoluteFilePath(realPath));
+    QUrl url = QUrl::fromLocalFile(QDir(documentPath()).filePath(realPath));
     if (!realFragment.isEmpty()) {
-        if (realFragment.startsWith("//apple_ref"))
+        if (realFragment.startsWith(QLatin1String("//apple_ref"))
+                || realFragment.startsWith(QLatin1String("//dash_ref"))) {
             url.setFragment(realFragment, QUrl::DecodedMode);
-        else
+        } else {
             url.setFragment(realFragment);
+        }
     }
 
     return url;
@@ -619,3 +652,260 @@ QString Docset::parseSymbolType(const QString &str)
 
     return aliases.value(str, str);
 }
+
+bool Docset::isFuzzySearchEnabled() const
+{
+    return m_fuzzySearchEnabled;
+}
+
+void Docset::setFuzzySearchEnabled(bool enabled)
+{
+    m_fuzzySearchEnabled = enabled;
+}
+
+/**
+ * \brief Returns score based on a substring position in a string.
+ * \param str Original string.
+ * \param index Index of the substring within \a str.
+ * \param length Substring length.
+ * \return Score value between 1 and 100.
+ */
+static int scoreFuzzy(const char *str, int index, int length)
+{
+    if (index == 0 || str[index - 1] == '.') {
+        // score between 66..99, if the match follows a dot, or starts the string
+        return qMax(66, 100 - length);
+    } else if (str[index + length] == 0) {
+        // score between 33..66, if the match is at the end of the string
+        return qMax(33, 67 - length);
+    } else {
+        // score between 1..33 otherwise (match in the middle of the string)
+        return qMax(1, 34 - length);
+    }
+}
+
+// Based on https://github.com/bevacqua/fuzzysearch
+static void matchFuzzy(const char *needle, int needleLength,
+                       const char *haystack, int haystackLength,
+                       int *start, int *length)
+{
+    static const int MaxDistance = 8;
+    static const int MaxGroupCount = 3;
+
+    *start = -1;
+
+    int groupCount = 0;
+    int bestRecursiveScore = -1;
+    int bestRecursiveStart = -1;
+    int bestRecursiveLength = -1;
+
+    for (int i = 0, j = 0; i < needleLength; ++i) {
+        bool found = false;
+        bool first = true;
+        int distance = 0;
+
+        while (j < haystackLength) {
+            if (needle[i] == haystack[j++]) {
+                if (*start == -1) {
+                    *start = j;  // first matched char
+
+                    // try starting the search later in case the first character occurs again later
+                    int recursiveStart;
+                    int recursiveLength;
+                    matchFuzzy(needle, needleLength, haystack + j,
+                               haystackLength - j,
+                               &recursiveStart, &recursiveLength);
+                    if (recursiveStart != -1) {
+                        int recursiveScore = scoreFuzzy(haystack,
+                                                        recursiveStart,
+                                                        recursiveLength);
+                        if (recursiveScore > bestRecursiveScore) {
+                            bestRecursiveScore = recursiveScore;
+                            bestRecursiveStart = recursiveStart;
+                            bestRecursiveLength = recursiveLength;
+                        }
+                    }
+                }
+
+                *length = j - *start + 1;
+                found = true;
+                break;
+            }
+
+            // Optimizations to reduce returned number of results
+            // (search was returning too many irrelevant results with large docsets)
+            // Optimization #1: too many mismatches.
+            if (first) {
+                if (++groupCount >= MaxGroupCount) {
+                    break;
+                }
+
+                first = false;
+            }
+
+            // Optimization #2: too large distance between found chars.
+            if (i != 0 && ++distance >= MaxDistance) {
+                break;
+            }
+        }
+
+        if (!found) {
+            // End of haystack, char not found.
+            if (bestRecursiveScore != -1) {
+                // Can still match with the same constraints if matching started later
+                // (smaller distance from first char to 2nd char)
+                *start = bestRecursiveStart;
+                *length = bestRecursiveLength;
+            } else {
+                *start = -1;
+            }
+            return;
+        }
+    }
+
+    int score = scoreFuzzy(haystack, *start, *length);
+    if (bestRecursiveScore > score) {
+        *start = bestRecursiveStart;
+        *length = bestRecursiveLength;
+    }
+}
+
+// Ported from DevDocs (https://github.com/Thibaut/devdocs), see app/searcher.coffee.
+static int scoreExact(int matchIndex, int matchLen, const char *value, int valueLen)
+{
+    static const char DOT = '.';
+
+    int score = 100;
+
+    // Remove one point for each unmatched character.
+    score -= (valueLen - matchLen);
+
+    if (matchIndex > 0) {
+        if (value[matchIndex - 1] == DOT) {
+            // If the character preceding the query is a dot, assign the same
+            // score as if the query was found at the beginning of the string,
+            // minus one.
+            score += matchIndex - 1;
+        } else if (matchLen == 1) {
+            // Don't match a single-character query unless it's found at the
+            // beginning of the string or is preceded by a dot.
+            return 0;
+        } else {
+            // (1) Remove one point for each unmatched character up to
+            //     the nearest preceding dot or the beginning of the
+            //     string.
+            // (2) Remove one point for each unmatched character
+            //     following the query.
+            int i = matchIndex - 2;
+            while (i >= 0 && value[i] != DOT)
+                --i;
+
+            score -= (matchIndex - i)                      // (1)
+                    + (valueLen - matchLen - matchIndex);  // (2)
+        }
+
+        // Remove one point for each dot preceding the query, except for the
+        // one immediately before the query.
+        for (int i = matchIndex - 2; i >= 0; --i) {
+            if (value[i] == DOT)
+                --score;
+        }
+    }
+
+    // Remove five points for each dot following the query.
+    for (int i = valueLen - matchLen - matchIndex - 1; i >= 0; --i) {
+        if (value[matchIndex + matchLen + i] == DOT)
+            score -= 5;
+    }
+
+    return qMax(1, score);
+}
+
+static inline int scoreFunction(const char *needleOrig, const char *haystackOrig)
+{
+    const int needleLength = static_cast<int>(qstrlen(needleOrig));
+    const int haystackLength = static_cast<int>(qstrlen(haystackOrig));
+
+    QVarLengthArray<char, 1024> needle(needleLength + 1);
+    QVarLengthArray<char, 1024> haystack(haystackLength + 1);
+
+    for (int i = 0, j = 0; i <= needleLength; ++i, ++j) {
+        const char c = needleOrig[i];
+        if ((i > 0 && needleOrig[i - 1] == ':' && c == ':') // C++ (::)
+                || c == '/' || c == '_' || c == ' ') { // Go, some Guides
+            needle[j] = '.';
+        } else if (c >= 'A' && c <= 'Z')  {
+            needle[j] = c + 32;
+        } else {
+            needle[j] = c;
+        }
+    }
+
+    for (int i = 0, j = 0; i <= haystackLength; ++i, ++j) {
+        const char c = haystackOrig[i];
+        if ((i > 0 && haystackOrig[i - 1] == ':' && c == ':') // C++ (::)
+                || c == '/' || c == '_' || c == ' ') { // Go, some Guides
+            haystack[j] = '.';
+        } else if (c >= 'A' && c <= 'Z')  {
+            haystack[j] = c + 32;
+        } else {
+            haystack[j] = c;
+        }
+    }
+
+    int score = 0;
+    int matchIndex = -1;
+    int matchLength = 0;
+    int exactIndex = -1;
+    const char *exactMatch = std::strstr(haystack.data(), needle.data());
+
+    if (exactMatch != nullptr) {
+        exactIndex = exactMatch - haystack.data();
+    }
+
+    if (exactIndex == -1) {
+        matchFuzzy(needle.data(), needleLength,
+                   haystack.data(), haystackLength,
+                   &matchIndex, &matchLength);
+    }
+
+    if (matchIndex == -1 && exactIndex == -1) { // no match
+        // simply return 0
+        return 0;
+    } else if (exactIndex != -1) {
+        // +100 to make sure exact matches are always on top.
+        score = scoreExact(exactIndex, needleLength, haystack.data(), haystackLength) + 100;
+    } else {
+        score = scoreFuzzy(haystack.data(), matchIndex, matchLength);
+
+        int indexOfLastDot;
+        for (indexOfLastDot = haystackLength - 1; indexOfLastDot >= 0; --indexOfLastDot) {
+            if (haystack[indexOfLastDot] == '.')
+                break;
+        }
+
+        if (indexOfLastDot != -1) {
+            matchIndex = -1;
+            matchFuzzy(needle.data(), needleLength,
+                       haystack.data() + indexOfLastDot + 1, haystackLength - (indexOfLastDot + 1),
+                       &matchIndex, &matchLength);
+
+            if (matchIndex != -1) {
+                score = qMax(score, scoreFuzzy(haystack.data() + indexOfLastDot + 1,
+                                               matchIndex, matchLength));
+            }
+        }
+    }
+
+    return score;
+}
+
+static void sqliteScoreFunction(sqlite3_context *context, int argc, sqlite3_value **argv)
+{
+    Q_UNUSED(argc);
+
+    const char *needle = reinterpret_cast<const char *>(sqlite3_value_text(argv[0]));
+    const char *haystack = reinterpret_cast<const char *>(sqlite3_value_text(argv[1]));
+
+    sqlite3_result_int(context, scoreFunction(needle, haystack));
+}
