@@ -29,7 +29,9 @@
 
 #include <core/application.h>
 #include <core/settings.h>
+#include <registry/docset.h>
 #include <registry/docsetregistry.h>
+#include <registry/itemdatarole.h>
 #include <registry/listmodel.h>
 
 #include <QClipboard>
@@ -45,6 +47,7 @@
 #include <QtConcurrent/QtConcurrent>
 
 using namespace Zeal;
+using namespace Zeal::WidgetUi;
 
 namespace {
 const char ApiServerUrl[] = "http://api.zealdocs.org/v1";
@@ -86,7 +89,7 @@ DocsetsDialog::DocsetsDialog(Core::Application *app, QWidget *parent) :
     ui->installedDocsetList->setItemDelegate(new DocsetListItemDelegate(this));
     ui->installedDocsetList->setModel(new ListModel(app->docsetRegistry(), this));
     connect(ui->installedDocsetList, &QListView::activated, this, [this](const QModelIndex &index) {
-        if (!index.data(ListModel::UpdateAvailableRole).toBool()) {
+        if (!index.data(Registry::ItemDataRole::UpdateAvailableRole).toBool()) {
             return;
         }
 
@@ -99,7 +102,7 @@ DocsetsDialog::DocsetsDialog(Core::Application *app, QWidget *parent) :
         ui->removeDocsetsButton->setEnabled(selectionModel->hasSelection());
 
         for (const QModelIndex &index : selectionModel->selectedRows()) {
-            if (index.data(ListModel::UpdateAvailableRole).toBool()) {
+            if (index.data(Registry::ItemDataRole::UpdateAvailableRole).toBool()) {
                 ui->updateSelectedDocsetsButton->setEnabled(true);
                 return;
             }
@@ -212,13 +215,12 @@ void DocsetsDialog::addDashFeed()
 
     QNetworkReply *reply = download(QUrl(feedUrl));
     reply->setProperty(DownloadTypeProperty, DownloadDashFeed);
-    connect(reply, &QNetworkReply::finished, this, &DocsetsDialog::downloadCompleted);
 }
 
 void DocsetsDialog::updateSelectedDocsets()
 {
     for (const QModelIndex &index : ui->installedDocsetList->selectionModel()->selectedRows()) {
-        if (!index.data(Registry::ListModel::UpdateAvailableRole).toBool())
+        if (!index.data(Registry::ItemDataRole::UpdateAvailableRole).toBool())
             continue;
 
         downloadDashDocset(index);
@@ -230,7 +232,7 @@ void DocsetsDialog::updateAllDocsets()
     QAbstractItemModel *model = ui->installedDocsetList->model();
     for (int i = 0; i < model->rowCount(); ++i) {
         const QModelIndex index = model->index(i, 0);
-        if (!index.data(Registry::ListModel::UpdateAvailableRole).toBool())
+        if (!index.data(Registry::ItemDataRole::UpdateAvailableRole).toBool())
             continue;
 
         downloadDashDocset(index);
@@ -262,7 +264,7 @@ void DocsetsDialog::removeSelectedDocsets()
     // Gather names first, because model indicies become invalid when docsets are removed.
     QStringList names;
     for (const QModelIndex &index : selectedIndexes)
-        names.append(index.data(Registry::ListModel::DocsetNameRole).toString());
+        names.append(index.data(Registry::ItemDataRole::DocsetNameRole).toString());
 
     for (const QString &name : names)
         removeDocset(name);
@@ -276,7 +278,7 @@ void DocsetsDialog::updateDocsetFilter(const QString &filterString)
         QListWidgetItem *item = ui->availableDocsetList->item(i);
 
         // Skip installed docsets
-        if (m_docsetRegistry->contains(item->data(Registry::ListModel::DocsetNameRole).toString()))
+        if (m_docsetRegistry->contains(item->data(Registry::ItemDataRole::DocsetNameRole).toString()))
             continue;
 
         item->setHidden(doSearch && !item->text().contains(filterString, Qt::CaseInsensitive));
@@ -329,9 +331,6 @@ void DocsetsDialog::downloadCompleted()
                 newReply->setProperty(DownloadTypeProperty, reply->property(DownloadTypeProperty));
                 newReply->setProperty(ListItemIndexProperty,
                                       reply->property(ListItemIndexProperty));
-
-                connect(newReply, &QNetworkReply::finished,
-                        this, &DocsetsDialog::downloadCompleted);
                 return;
             }
 
@@ -362,8 +361,6 @@ void DocsetsDialog::downloadCompleted()
         newReply->setProperty(DownloadTypeProperty, reply->property(DownloadTypeProperty));
         newReply->setProperty(ListItemIndexProperty, reply->property(ListItemIndexProperty));
 
-        connect(newReply, &QNetworkReply::finished, this, &DocsetsDialog::downloadCompleted);
-
         return;
     }
 
@@ -382,8 +379,16 @@ void DocsetsDialog::downloadCompleted()
         const QJsonDocument jsonDoc = QJsonDocument::fromJson(data, &jsonError);
 
         if (jsonError.error != QJsonParseError::NoError) {
-            QMessageBox::warning(this, QStringLiteral("Zeal"),
-                                 tr("Corrupted docset list: ") + jsonError.errorString());
+            // TODO: Log QJsonParseError.
+            const QMessageBox::StandardButton rc
+                    = QMessageBox::warning(this, QStringLiteral("Zeal"),
+                                           tr("Server returned a corrupted docset list."),
+                                           QMessageBox::Retry | QMessageBox::Cancel);
+
+            if (rc == QMessageBox::Retry) {
+                downloadDocsetList();
+            }
+
             break;
         }
 
@@ -404,7 +409,7 @@ void DocsetsDialog::downloadCompleted()
         QNetworkReply *reply = download(metadata.url());
         reply->setProperty(DocsetNameProperty, metadata.name());
         reply->setProperty(DownloadTypeProperty, DownloadDocset);
-        connect(reply, &QNetworkReply::finished, this, &DocsetsDialog::downloadCompleted);
+
         break;
     }
 
@@ -492,7 +497,7 @@ void DocsetsDialog::extractionCompleted(const QString &filePath)
     const QString docsetName = docsetNameForTmpFilePath(filePath);
 
     const QDir dataDir(m_application->settings()->docsetPath);
-    const QString docsetPath = dataDir.absoluteFilePath(docsetName + QLatin1String(".docset"));
+    const QString docsetPath = dataDir.filePath(docsetName + QLatin1String(".docset"));
 
     // Write metadata about docset
     Registry::DocsetMetadata metadata = m_availableDocsets.contains(docsetName)
@@ -566,7 +571,7 @@ QListWidgetItem *DocsetsDialog::findDocsetListItem(const QString &name) const
     for (int i = 0; i < ui->availableDocsetList->count(); ++i) {
         QListWidgetItem *item = ui->availableDocsetList->item(i);
 
-        if (item->data(Registry::ListModel::DocsetNameRole).toString() == name)
+        if (item->data(Registry::ItemDataRole::DocsetNameRole).toString() == name)
             return item;
     }
 
@@ -587,6 +592,7 @@ QNetworkReply *DocsetsDialog::download(const QUrl &url)
 {
     QNetworkReply *reply = m_application->download(url);
     connect(reply, &QNetworkReply::downloadProgress, this, &DocsetsDialog::downloadProgress);
+    connect(reply, &QNetworkReply::finished, this, &DocsetsDialog::downloadCompleted);
     m_replies.append(reply);
 
     // Installed docsets
@@ -628,7 +634,6 @@ void DocsetsDialog::downloadDocsetList()
 
     QNetworkReply *reply = download(QUrl(ApiServerUrl + QLatin1String("/docsets")));
     reply->setProperty(DownloadTypeProperty, DownloadDocsetList);
-    connect(reply, &QNetworkReply::finished, this, &DocsetsDialog::downloadCompleted);
 }
 
 void DocsetsDialog::processDocsetList(const QJsonArray &list)
@@ -644,7 +649,7 @@ void DocsetsDialog::processDocsetList(const QJsonArray &list)
     for (const Registry::DocsetMetadata &metadata : m_availableDocsets) {
         QListWidgetItem *listItem
                 = new QListWidgetItem(metadata.icon(), metadata.title(), ui->availableDocsetList);
-        listItem->setData(Registry::ListModel::DocsetNameRole, metadata.name());
+        listItem->setData(Registry::ItemDataRole::DocsetNameRole, metadata.name());
 
         if (m_docsetRegistry->contains(metadata.name())) {
             listItem->setHidden(true);
@@ -652,8 +657,7 @@ void DocsetsDialog::processDocsetList(const QJsonArray &list)
             Registry::Docset *docset = m_docsetRegistry->docset(metadata.name());
 
             if (metadata.latestVersion() != docset->version()
-                    || (metadata.latestVersion() == docset->version()
-                        && metadata.revision() > docset->revision())) {
+                    || metadata.revision() > docset->revision()) {
                 docset->hasUpdate = true;
                 ui->updateAllDocsetsButton->setEnabled(true);
             }
@@ -665,7 +669,7 @@ void DocsetsDialog::processDocsetList(const QJsonArray &list)
 
 void DocsetsDialog::downloadDashDocset(const QModelIndex &index)
 {
-    const QString name = index.data(Registry::ListModel::DocsetNameRole).toString();
+    const QString name = index.data(Registry::ItemDataRole::DocsetNameRole).toString();
 
     if (!m_availableDocsets.contains(name))
         return;
@@ -676,8 +680,6 @@ void DocsetsDialog::downloadDashDocset(const QModelIndex &index)
     reply->setProperty(DownloadTypeProperty, DownloadDocset);
     reply->setProperty(ListItemIndexProperty,
                        ui->availableDocsetList->row(findDocsetListItem(name)));
-
-    connect(reply, &QNetworkReply::finished, this, &DocsetsDialog::downloadCompleted);
 }
 
 void DocsetsDialog::removeDocset(const QString &name)
@@ -757,7 +759,7 @@ void DocsetsDialog::resetProgress()
     QItemSelectionModel *selectionModel = ui->installedDocsetList->selectionModel();
     bool hasSelectedUpdates = false;
     for (const QModelIndex &index : selectionModel->selectedRows()) {
-        if (index.data(Registry::ListModel::UpdateAvailableRole).toBool()) {
+        if (index.data(Registry::ItemDataRole::UpdateAvailableRole).toBool()) {
             hasSelectedUpdates = true;
             break;
         }
@@ -773,9 +775,9 @@ void DocsetsDialog::resetProgress()
 
 QString DocsetsDialog::docsetNameForTmpFilePath(const QString &filePath) const
 {
-    for (const QString &key : m_tmpFiles.keys()) {
-        if (m_tmpFiles[key]->fileName() == filePath) {
-            return key;
+    for (auto it = m_tmpFiles.cbegin(), end = m_tmpFiles.cend(); it != end; ++it) {
+        if (it.value()->fileName() == filePath) {
+            return it.key();
         }
     }
 
