@@ -32,12 +32,15 @@
 
 #include <core/application.h>
 #include <core/settings.h>
+#include <registry/docset.h>
 #include <registry/docsetregistry.h>
+#include <registry/itemdatarole.h>
 #include <registry/listmodel.h>
 #include <registry/searchmodel.h>
 
 #include <QCloseEvent>
 #include <QDesktopServices>
+#include <QFileInfo>
 #include <QKeyEvent>
 #include <QMenu>
 #include <QMessageBox>
@@ -46,35 +49,23 @@
 #include <QSystemTrayIcon>
 #include <QTabBar>
 #include <QTimer>
-
-#ifdef USE_WEBENGINE
-#include <QWebEngineHistory>
-#include <QWebEngineHistoryItem>
-#include <QWebEnginePage>
-
-typedef QWebEngineHistory QWebHistory;
-typedef QWebEngineHistoryItem QWebHistoryItem;
-typedef QWebEnginePage QWebPage;
-#else
 #include <QWebFrame>
 #include <QWebHistory>
 #include <QWebPage>
-#endif
-
-// TODO: [Qt 5.5] Remove in favour of native Qt support (QTBUG-31762)
-#ifdef USE_APPINDICATOR
-#undef signals
-#include <libappindicator/app-indicator.h>
-#define signals public
-#include <gtk/gtk.h>
-#endif
 
 using namespace Zeal;
+using namespace Zeal::WidgetUi;
 
 namespace {
-const char startPageUrl[] = "qrc:///browser/start.html";
+const char WelcomePageUrl[] = "qrc:///browser/welcome.html";
+const char WelcomePageNoAdUrl[] = "qrc:///browser/welcome-noad.html";
+const char DarkModeCssUrl[] = ":/browser/assets/css/darkmode.css";
+const char HighlightOnNavigateCssUrl[] = ":/browser/assets/css/highlight.css";
 }
 
+namespace Zeal {
+namespace WidgetUi {
+
 struct TabState
 {
     explicit TabState()
@@ -83,10 +74,8 @@ struct TabState
         tocModel = new Registry::SearchModel();
 
         webPage = new QWebPage();
-#ifndef USE_WEBENGINE
         webPage->setLinkDelegationPolicy(QWebPage::DelegateExternalLinks);
         webPage->setNetworkAccessManager(Core::Application::instance()->networkManager());
-#endif
     }
 
     TabState(const TabState &other)
@@ -101,10 +90,8 @@ struct TabState
         tocModel = new Registry::SearchModel(*other.tocModel);
 
         webPage = new QWebPage();
-#ifndef USE_WEBENGINE
         webPage->setLinkDelegationPolicy(QWebPage::DelegateExternalLinks);
         webPage->setNetworkAccessManager(Core::Application::instance()->networkManager());
-#endif
 
         restoreHistory(other.saveHistory());
     }
@@ -132,29 +119,17 @@ struct TabState
     }
 
     QUrl url() const {
-#ifdef USE_WEBENGINE
-        return webPage->url();
-#else
         return webPage->mainFrame()->url();
-#endif
     }
 
     void loadUrl(const QUrl &url)
     {
-#ifdef USE_WEBENGINE
-        webPage->load(url);
-#else
         webPage->mainFrame()->load(url);
-#endif
     }
 
     QString title() const
     {
-#ifdef USE_WEBENGINE
-        return webPage->title();
-#else
         return webPage->mainFrame()->title();
-#endif
     }
 
     QString searchQuery;
@@ -173,6 +148,9 @@ struct TabState
     int webViewZoomFactor = 0;
 };
 
+} // namespace WidgetUi
+} // namespace Zeal
+
 MainWindow::MainWindow(Core::Application *app, QWidget *parent) :
     QMainWindow(parent),
     ui(new Ui::MainWindow),
@@ -184,15 +162,6 @@ MainWindow::MainWindow(Core::Application *app, QWidget *parent) :
 {
     ui->setupUi(this);
 
-    connect(m_settings, &Core::Settings::updated, this, &MainWindow::applySettings);
-
-#ifdef USE_APPINDICATOR
-    detectAppIndicatorSupport();
-#endif
-
-    if (m_settings->showSystrayIcon)
-        createTrayIcon();
-
     // initialise key grabber
     connect(m_globalShortcut, &QxtGlobalShortcut::activated, this, &MainWindow::toggleWindow);
 
@@ -317,7 +286,7 @@ MainWindow::MainWindow(Core::Application *app, QWidget *parent) :
     ui->lineEdit->setFocus();
     setupSearchBoxCompletions();
     SearchItemDelegate *delegate = new SearchItemDelegate(ui->treeView);
-    delegate->setDecorationRoles({Registry::SearchModel::DocsetIconRole, Qt::DecorationRole});
+    delegate->setDecorationRoles({Registry::ItemDataRole::DocsetIconRole, Qt::DecorationRole});
     connect(ui->lineEdit, &QLineEdit::textChanged, [delegate](const QString &text) {
         delegate->setHighlight(Registry::SearchQuery::fromString(text).query());
     });
@@ -363,32 +332,39 @@ MainWindow::MainWindow(Core::Application *app, QWidget *parent) :
             QDesktopServices::openUrl(url);
     });
 
-    connect(m_application->docsetRegistry(), &Registry::DocsetRegistry::queryCompleted,
+    connect(m_application->docsetRegistry(), &Registry::DocsetRegistry::searchCompleted,
             this, [this](const QList<Registry::SearchResult> &results) {
         currentTabState()->searchModel->setResults(results);
     });
 
-    connect(m_application->docsetRegistry(), &Registry::DocsetRegistry::docsetRemoved,
+    connect(m_application->docsetRegistry(), &Registry::DocsetRegistry::docsetAboutToBeRemoved,
             this, [this](const QString &name) {
-        setupSearchBoxCompletions();
         for (TabState *tabState : m_tabStates) {
-            if (docsetName(tabState->url()) != name)
-                continue;
+            if (tabState == currentTabState()) {
+                // Disable updates because removeSearchResultWithName can
+                // call {begin,end}RemoveRows multiple times, and cause
+                // degradation of UI responsiveness.
+                ui->treeView->setUpdatesEnabled(false);
+                tabState->searchModel->removeSearchResultWithName(name);
+                ui->treeView->setUpdatesEnabled(true);
+            } else {
+                tabState->searchModel->removeSearchResultWithName(name);
+            }
 
-            tabState->tocModel->setResults();
-
-            // optimization: disable updates temporarily because
-            // removeSearchResultWithName can call {begin,end}RemoveRows
-            // multiple times which can cause GUI updates to be suboptimal
-            // in case of many rows to be removed
-            ui->treeView->setUpdatesEnabled(false);
-            tabState->searchModel->removeSearchResultWithName(name);
-            ui->treeView->setUpdatesEnabled(true);
-
-            tabState->loadUrl(QUrl(startPageUrl));
+            if (docsetName(tabState->url()) == name) {
+                tabState->tocModel->setResults();
+                // TODO: Add custom 'Page has been removed' page.
+                if (m_settings->isAdDisabled) {
+                    tabState->loadUrl(QUrl(WelcomePageNoAdUrl));
+                } else {
+                    tabState->loadUrl(QUrl(WelcomePageUrl));
+                }
+            }
 
             // TODO: Cleanup history
         }
+
+        setupSearchBoxCompletions();
     });
 
     connect(m_application->docsetRegistry(), &Registry::DocsetRegistry::docsetAdded,
@@ -401,13 +377,7 @@ MainWindow::MainWindow(Core::Application *app, QWidget *parent) :
             return;
 
         currentTabState()->searchQuery = text;
-        m_cancelSearch.cancel();
-        m_cancelSearch = Registry::CancellationToken();
-        m_application->docsetRegistry()->search(text, m_cancelSearch);
-        if (text.isEmpty()) {
-            currentTabState()->tocModel->setResults();
-            syncTreeView();
-        }
+        m_application->docsetRegistry()->search(text);
     });
 
     // Setup delayed navigation to a page until user makes a pause in typing a search query.
@@ -420,7 +390,7 @@ MainWindow::MainWindow(Core::Application *app, QWidget *parent) :
 
         openDocset(index);
 
-        // Get focus back. QWebPageEngine::load() always steals focus.
+        // Get focus back.
         ui->lineEdit->setFocus(Qt::MouseFocusReason);
     });
 
@@ -452,14 +422,16 @@ MainWindow::MainWindow(Core::Application *app, QWidget *parent) :
             QDesktopServices::openUrl(url);
     });
 
-    ui->actionNextTab->setShortcut(QKeySequence::NextChild);
+    ui->actionNextTab->setShortcuts({QKeySequence::NextChild,
+                                     QKeySequence(Qt::ControlModifier| Qt::Key_PageDown)});
     addAction(ui->actionNextTab);
     connect(ui->actionNextTab, &QAction::triggered, [this]() {
         m_tabBar->setCurrentIndex((m_tabBar->currentIndex() + 1) % m_tabBar->count());
     });
 
     // TODO: Use QKeySequence::PreviousChild, when QTBUG-15746 is fixed.
-    ui->actionPreviousTab->setShortcut(Qt::ControlModifier | Qt::ShiftModifier | Qt::Key_Tab);
+    ui->actionPreviousTab->setShortcuts({QKeySequence(Qt::ControlModifier | Qt::ShiftModifier | Qt::Key_Tab),
+                                         QKeySequence(Qt::ControlModifier| Qt::Key_PageUp)});
     addAction(ui->actionPreviousTab);
     connect(ui->actionPreviousTab, &QAction::triggered, [this]() {
         m_tabBar->setCurrentIndex((m_tabBar->currentIndex() - 1 + m_tabBar->count()) % m_tabBar->count());
@@ -470,6 +442,9 @@ MainWindow::MainWindow(Core::Application *app, QWidget *parent) :
     ui->tocListView->setAttribute(Qt::WA_MacShowFocusRect, false);
 #endif
 
+    connect(m_settings, &Core::Settings::updated, this, &MainWindow::applySettings);
+    applySettings();
+
     if (m_settings->checkForUpdate)
         m_application->checkForUpdates(true);
 }
@@ -490,21 +465,17 @@ void MainWindow::search(const Registry::SearchQuery &query)
         return;
 
     ui->lineEdit->setText(query.toString());
-    ui->treeView->activated(ui->treeView->currentIndex());
+    emit ui->treeView->activated(ui->treeView->currentIndex());
 }
 
 void MainWindow::openDocset(const QModelIndex &index)
 {
-    const QVariant url = index.sibling(index.row(), 1).data();
+    const QVariant url = index.data(Registry::ItemDataRole::UrlRole);
     if (url.isNull())
         return;
 
     ui->webView->load(url.toUrl());
-
-    // QWebEnginePage::load() always steals focus, so no need to do it twice.
-#ifndef USE_WEBENGINE
     ui->webView->focus();
-#endif
 }
 
 QString MainWindow::docsetName(const QUrl &url) const
@@ -556,14 +527,20 @@ void MainWindow::createTab(int index)
 
     using Registry::SearchModel;
     TabState *newTab = new TabState();
-    connect(newTab->searchModel, &SearchModel::queryCompleted, this, &MainWindow::queryCompleted);
-    connect(newTab->tocModel, &SearchModel::queryCompleted, this, &MainWindow::syncToc);
+    connect(newTab->searchModel, &SearchModel::updated, this, &MainWindow::queryCompleted);
+    connect(newTab->tocModel, &SearchModel::updated, this, &MainWindow::syncToc);
 
-    newTab->loadUrl(QUrl(startPageUrl));
+    if (m_settings->isAdDisabled) {
+        newTab->loadUrl(QUrl(WelcomePageNoAdUrl));
+    } else {
+        newTab->loadUrl(QUrl(WelcomePageUrl));
+    }
 
     m_tabStates.insert(index, newTab);
     m_tabBar->insertTab(index, tr("Loading..."));
     m_tabBar->setCurrentIndex(index);
+
+    ui->lineEdit->setFocus();
 }
 
 void MainWindow::duplicateTab(int index)
@@ -573,8 +550,8 @@ void MainWindow::duplicateTab(int index)
 
     using Registry::SearchModel;
     TabState *newTab = new TabState(*m_tabStates.at(index));
-    connect(newTab->searchModel, &SearchModel::queryCompleted, this, &MainWindow::queryCompleted);
-    connect(newTab->tocModel, &SearchModel::queryCompleted, this, &MainWindow::syncToc);
+    connect(newTab->searchModel, &SearchModel::updated, this, &MainWindow::queryCompleted);
+    connect(newTab->tocModel, &SearchModel::updated, this, &MainWindow::syncToc);
 
     ++index;
     m_tabStates.insert(index, newTab);
@@ -712,118 +689,39 @@ void MainWindow::setupTabBar()
     layout->insertWidget(2, m_tabBar, 0, Qt::AlignBottom);
 }
 
-#ifdef USE_APPINDICATOR
-void MainWindow::detectAppIndicatorSupport()
-{
-    const QByteArray xdgDesktop = qgetenv("XDG_CURRENT_DESKTOP");
-
-    // Unity
-    if (xdgDesktop == "Unity") {
-        m_useAppIndicator = true;
-        return;
-    }
-
-    // Cinnamon 2.8
-    // Checking specifically for 2.8 because direct AppIndicator support will be dropped soon.
-    if (xdgDesktop == "X-Cinnamon" && qgetenv("CINNAMON_VERSION").startsWith("2.8")) {
-        m_useAppIndicator = true;
-        return;
-    }
-}
-#endif
-
-#ifdef USE_APPINDICATOR
-void appIndicatorToggleWindow(GtkMenu *menu, gpointer data)
-{
-    Q_UNUSED(menu)
-    static_cast<MainWindow *>(data)->toggleWindow();
-}
-#endif
-
 void MainWindow::createTrayIcon()
 {
-#ifdef USE_APPINDICATOR
-    if (m_trayIcon || m_appIndicator)
-        return;
-#else
     if (m_trayIcon)
         return;
-#endif
 
-#ifdef USE_APPINDICATOR
-    if (m_useAppIndicator) {
-        m_appIndicatorMenu = gtk_menu_new();
+    m_trayIcon = new QSystemTrayIcon(this);
+    m_trayIcon->setIcon(windowIcon());
+    m_trayIcon->setToolTip(QStringLiteral("Zeal"));
 
-        m_appIndicatorShowHideMenuItem = gtk_menu_item_new_with_label(qPrintable(tr("Hide")));
-        gtk_menu_shell_append(GTK_MENU_SHELL(m_appIndicatorMenu), m_appIndicatorShowHideMenuItem);
-        g_signal_connect(m_appIndicatorShowHideMenuItem, "activate",
-                         G_CALLBACK(appIndicatorToggleWindow), this);
+    connect(m_trayIcon, &QSystemTrayIcon::activated, [this](QSystemTrayIcon::ActivationReason reason) {
+        if (reason != QSystemTrayIcon::Trigger && reason != QSystemTrayIcon::DoubleClick)
+            return;
 
-        m_appIndicatorMenuSeparator = gtk_separator_menu_item_new();
-        gtk_menu_shell_append(GTK_MENU_SHELL(m_appIndicatorMenu), m_appIndicatorMenuSeparator);
+        toggleWindow();
+    });
 
-        m_appIndicatorQuitMenuItem = gtk_menu_item_new_with_label(qPrintable(tr("Quit")));
-        gtk_menu_shell_append(GTK_MENU_SHELL(m_appIndicatorMenu), m_appIndicatorQuitMenuItem);
-        g_signal_connect(m_appIndicatorQuitMenuItem, "activate",
-                         G_CALLBACK(QCoreApplication::quit), NULL);
+    QMenu *trayIconMenu = new QMenu(this);
+    trayIconMenu->addAction(ui->actionQuit);
 
-        gtk_widget_show_all(m_appIndicatorMenu);
+    m_trayIcon->setContextMenu(trayIconMenu);
 
-        // NOTE: Zeal icon has to be installed, otherwise app indicator won't be shown
-        m_appIndicator = app_indicator_new("zeal", "zeal", APP_INDICATOR_CATEGORY_OTHER);
-
-        app_indicator_set_status(m_appIndicator, APP_INDICATOR_STATUS_ACTIVE);
-        app_indicator_set_menu(m_appIndicator, GTK_MENU(m_appIndicatorMenu));
-    } else {  // others
-#endif
-        m_trayIcon = new QSystemTrayIcon(this);
-        m_trayIcon->setIcon(windowIcon());
-        m_trayIcon->setToolTip(QStringLiteral("Zeal"));
-
-        connect(m_trayIcon, &QSystemTrayIcon::activated, [this](QSystemTrayIcon::ActivationReason reason) {
-            if (reason != QSystemTrayIcon::Trigger && reason != QSystemTrayIcon::DoubleClick)
-                return;
-
-            toggleWindow();
-        });
-
-        QMenu *trayIconMenu = new QMenu(this);
-        trayIconMenu->addAction(ui->actionQuit);
-
-        m_trayIcon->setContextMenu(trayIconMenu);
-
-        m_trayIcon->show();
-#ifdef USE_APPINDICATOR
-    }
-#endif
+    m_trayIcon->show();
 }
 
 void MainWindow::removeTrayIcon()
 {
-#ifdef USE_APPINDICATOR
-    if (!m_trayIcon && !m_appIndicator)
-        return;
-#else
     if (!m_trayIcon)
         return;
-#endif
 
-#ifdef USE_APPINDICATOR
-    if (m_useAppIndicator) {
-        g_clear_object(&m_appIndicator);
-        g_clear_object(&m_appIndicatorMenu);
-        g_clear_object(&m_appIndicatorShowHideMenuItem);
-        g_clear_object(&m_appIndicatorMenuSeparator);
-        g_clear_object(&m_appIndicatorQuitMenuItem);
-    } else {
-#endif
-        QMenu *trayIconMenu = m_trayIcon->contextMenu();
-        delete m_trayIcon;
-        m_trayIcon = nullptr;
-        delete trayIconMenu;
-#ifdef USE_APPINDICATOR
-    }
-#endif
+    QMenu *trayIconMenu = m_trayIcon->contextMenu();
+    delete m_trayIcon;
+    m_trayIcon = nullptr;
+    delete trayIconMenu;
 }
 
 void MainWindow::bringToFront()
@@ -905,6 +803,32 @@ void MainWindow::applySettings()
         createTrayIcon();
     else
         removeTrayIcon();
+
+    // Content
+    QByteArray ba;
+    if (m_settings->darkModeEnabled) {
+        QScopedPointer<QFile> file(new QFile(DarkModeCssUrl));
+        if (file->open(QIODevice::ReadOnly)) {
+            ba += file->readAll();
+        }
+    }
+
+    if (m_settings->highlightOnNavigateEnabled) {
+        QScopedPointer<QFile> file(new QFile(HighlightOnNavigateCssUrl));
+        if (file->open(QIODevice::ReadOnly)) {
+            ba += file->readAll();
+        }
+    }
+
+    if (QFileInfo::exists(m_settings->customCssFile)) {
+        QScopedPointer<QFile> file(new QFile(m_settings->customCssFile));
+        if (file->open(QIODevice::ReadOnly)) {
+            ba += file->readAll();
+        }
+    }
+
+    const QString cssUrl = QLatin1String("data:text/css;charset=utf-8;base64,") + ba.toBase64();
+    QWebSettings::globalSettings()->setUserStyleSheetUrl(QUrl(cssUrl));
 }
 
 void MainWindow::toggleWindow()
@@ -912,23 +836,9 @@ void MainWindow::toggleWindow()
     const bool checkActive = sender() == m_globalShortcut;
 
     if (!isVisible() || (checkActive && !isActiveWindow())) {
-#ifdef USE_APPINDICATOR
-        if (m_appIndicator) {
-            gtk_menu_item_set_label(GTK_MENU_ITEM(m_appIndicatorShowHideMenuItem),
-                                    qPrintable(tr("Hide")));
-        }
-#endif
         bringToFront();
     } else {
-#ifdef USE_APPINDICATOR
-        if (m_trayIcon || m_appIndicator) {
-            if (m_appIndicator) {
-                gtk_menu_item_set_label(GTK_MENU_ITEM(m_appIndicatorShowHideMenuItem),
-                                        qPrintable(tr("Show")));
-            }
-#else
         if (m_trayIcon) {
-#endif
             hide();
         } else {
             showMinimized();
