@@ -26,20 +26,13 @@
 
 #include <core/application.h>
 #include <core/settings.h>
-#include <registry/docsetregistry.h>
 
 #include <QDir>
 #include <QFileDialog>
-
-#ifdef USE_WEBENGINE
-#include <QWebEngineSettings>
-
-typedef QWebEngineSettings QWebSettings;
-#else
 #include <QWebSettings>
-#endif
 
 using namespace Zeal;
+using namespace Zeal::WidgetUi;
 
 SettingsDialog::SettingsDialog(Core::Application *app, QWidget *parent) :
     QDialog(parent),
@@ -57,7 +50,7 @@ SettingsDialog::SettingsDialog(Core::Application *app, QWidget *parent) :
             saveSettings();
     });
 
-    connect(ui->minFontSize, static_cast<void (QSpinBox::*)(int)>(&QSpinBox::valueChanged),
+    connect(ui->minimumFontSizeSpinBox, static_cast<void (QSpinBox::*)(int)>(&QSpinBox::valueChanged),
             this, [](int value) {
         QWebSettings::globalSettings()->setFontSize(QWebSettings::MinimumFontSize, value);
     });
@@ -70,13 +63,25 @@ SettingsDialog::~SettingsDialog()
     delete ui;
 }
 
-void SettingsDialog::on_storageButton_clicked()
+void SettingsDialog::chooseCustomCssFile()
+{
+    const QString file = QFileDialog::getOpenFileName(this, tr("Choose CSS File"),
+                                                      ui->customCssFileEdit->text(),
+                                                      tr("CSS Files (*.css);;All Files (*.*)"));
+    if (file.isEmpty())
+        return;
+
+    ui->customCssFileEdit->setText(QDir::toNativeSeparators(file));
+}
+
+void SettingsDialog::chooseDocsetStoragePath()
 {
     const QString dir = QFileDialog::getExistingDirectory(this, tr("Open Directory"),
-                                                          ui->storageEdit->text());
-    if (!dir.isEmpty())
-        ui->storageEdit->setText(QDir::toNativeSeparators(dir));
+                                                          ui->docsetStorageEdit->text());
+    if (dir.isEmpty())
+        return;
 
+    ui->docsetStorageEdit->setText(QDir::toNativeSeparators(dir));
 }
 
 void SettingsDialog::loadSettings()
@@ -92,11 +97,20 @@ void SettingsDialog::loadSettings()
 
     ui->toolButton->setKeySequence(settings->showShortcut);
 
+    ui->docsetStorageEdit->setText(QDir::toNativeSeparators(settings->docsetPath));
+
+    // Tabs Tab
     ui->openNewTabAfterActive->setChecked(settings->openNewTabAfterActive);
 
-    //
-    ui->minFontSize->setValue(settings->minimumFontSize);
-    ui->storageEdit->setText(QDir::toNativeSeparators(settings->docsetPath));
+    // Search Tab
+    ui->fuzzySearchCheckBox->setChecked(settings->fuzzySearchEnabled);
+
+    // Content Tab
+    ui->minimumFontSizeSpinBox->setValue(settings->minimumFontSize);
+    ui->darkModeCheckBox->setChecked(settings->darkModeEnabled);
+    ui->highlightOnNavigateCheckBox->setChecked(settings->highlightOnNavigateEnabled);
+    ui->customCssFileEdit->setText(QDir::toNativeSeparators(settings->customCssFile));
+    ui->disableAdCheckBox->setChecked(settings->isAdDisabled);
 
     // Network Tab
     switch (settings->proxyType) {
@@ -121,6 +135,7 @@ void SettingsDialog::loadSettings()
 void SettingsDialog::saveSettings()
 {
     Core::Settings * const settings = m_application->settings();
+
     // General Tab
     settings->startMinimized = ui->startMinimizedCheckBox->isChecked();
     settings->checkForUpdate = ui->checkForUpdateCheckBox->isChecked();
@@ -131,15 +146,20 @@ void SettingsDialog::saveSettings()
 
     settings->showShortcut = ui->toolButton->keySequence();
 
+    settings->docsetPath = QDir::fromNativeSeparators(ui->docsetStorageEdit->text());
+
+    // Tabs Tab
     settings->openNewTabAfterActive = ui->openNewTabAfterActive->isChecked();
 
-    //
-    settings->minimumFontSize = ui->minFontSize->text().toInt();
+    // Search Tab
+    settings->fuzzySearchEnabled = ui->fuzzySearchCheckBox->isChecked();
 
-    if (QDir::fromNativeSeparators(ui->storageEdit->text()) != settings->docsetPath) {
-        settings->docsetPath = QDir::fromNativeSeparators(ui->storageEdit->text());
-        m_application->docsetRegistry()->init(settings->docsetPath);
-    }
+    // Content Tab
+    settings->minimumFontSize = ui->minimumFontSizeSpinBox->text().toInt();
+    settings->darkModeEnabled = ui->darkModeCheckBox->isChecked();
+    settings->highlightOnNavigateEnabled = ui->highlightOnNavigateCheckBox->isChecked();
+    settings->customCssFile = QDir::fromNativeSeparators(ui->customCssFileEdit->text());
+    settings->isAdDisabled = ui->disableAdCheckBox->isChecked();
 
     // Network Tab
     // Proxy settings
